from .gan import GAN
from .vqvae import VQVAE
from .wgan import WGAN
from .auto_encoder import AutoEncoder
from .binary_classifier import BinaryClassifier

__all__ = [
    "GAN",
    "VQVAE",
    "WGAN",
    "AutoEncoder",
    "BinaryClassifier",
]